"""Whisper services."""

from . import base

import pathlib
from abc import ABC, abstractmethod

try:
    import whisper
except ImportError:
    raise ImportError("Please install whisper with `pip install whisper`")


class WhisperService(base.Service, ABC):
    @property
    def service_name(self) -> str:
        return "Whisper"


class WhisperSTTService(base.STTService, WhisperService):
    def __init__(
            self,
            model: str = "base",
            device: str | None = None,
            *,
            cache_dir: pathlib.Path | str | None = None
    ) -> None:
        super().__init__(cache_dir=cache_dir)
        self.model = model
        self.device = device

        self.model_obj = whisper.load_model(self.model, device=self.device)

    def stt(self, input: base.STTInput) -> base.STTData:
        info = base.ServiceInfo(
            service_name=self.service_name,
            service_type=self.service_type,
            config={"model": self.model, "device": self.device}
        )

        transcript = self.model_obj.transcribe(str(self.cache_dir / input.audio_path), word_timestamps=True)

        word_boundaries: list[base.Boundary] = []
        text_offset = 0
        for segment in transcript["segments"]:
            for word in segment["words"]:
                word_boundaries.append(
                    base.Boundary(
                        text=word["word"],
                        start=word["start"],
                        end=word["end"],
                        text_offset=text_offset
                    )
                )
                if text_offset != 0 and response.text[text_offset] == " ":
                    text_offset += 1
                text_offset += len(word["word"])
        
        return base.STTData(info=info, input=input, output=base.STTOutput(text=transcript["text"], boundaries=word_boundaries))




"""manim.Scene subclass with speech."""

from . import services
from . import speech
from . import translation

from collections import abc
import contextlib
import pathlib
import gettext

import manim


class VoiceoverScene(manim.Scene):
    tts_service: services.base.TTSService | None = None
    stt_service: services.base.STTService | None = None
    current_speech_data: speech.SpeechData | None = None
    current_speech_start_time: float | None = None

    def set_tts_service(self, service: services.base.TTSService) -> None:
        self.tts_service = service
    
    def set_stt_service(self, service: services.base.STTService) -> None:
        self.stt_service = service
    
    def safe_wait(self, duration: float) -> None:
        if duration > 1 / manim.config.frame_rate:
            self.wait(duration)
    
    def wait_for_voiceover(self) -> None:
        if not isinstance(self.current_speech_data, type(None)) and not isinstance(self.current_speech_start_time, type(None)):
            self.safe_wait(self.current_speech_data.duration - (self.renderer.time - self.current_speech_start_time))
    
    def wait_until_bookmark(self, key: str) -> None:
        if not isinstance(self.current_speech_data, type(None)) and not isinstance(self.current_speech_start_time, type(None)):
            self.safe_wait(self.current_speech_data.bookmark_times[key] - (self.renderer.time - self.current_speech_start_time))
    
    @contextlib.contextmanager
    def voiceover(self, text: str) -> abc.Generator[speech.SpeechData, None, None]:
        if not isinstance(self.stt_service, services.base.STTService):
            manim.logger.warning("No STT service is set. Bookmark locations will be inaccurate.")
        try:
            self.current_speech_data = speech.create(text, self.tts_service, self.stt_service)
            self.current_speech_start_time = self.renderer.time
            self.add_sound(str(pathlib.Path(manim.config.media_dir) / "manim_speech" / self.current_speech_data.tts_data.output.audio_path))
            yield self.current_speech_data
        finally:
            self.wait_for_voiceover()
            self.current_speech_data = None
            self.current_speech_start_time = None


class TranslationScene(manim.Scene):
    translation_service: services.base.TranslationService | None = None
    _ = staticmethod(gettext.gettext)

    def set_translation_service(self, service: services.base.TranslationService) -> None:
        self.translation_service = service
    
    def translate(self, file: pathlib.Path | str, domain: str, source_language: str, target_language: str,) -> None:
        translation.init_translation_env(file, domain)
        translation.translate_po_file(domain, source_language, target_language, service=self.translation_service)
        trans = gettext.translation(domain, languages=[target_language], localedir="locales")
        self._ = trans.gettext



"""OpenAI services."""

from . import base

import os
import pathlib
from abc import ABC, abstractmethod

try:
    import openai
except ImportError:
    raise ImportError("Please install openai with `pip install openai`")


class OpenAIService(base.Service, ABC):
    def __init__(self, *, cache_dir: pathlib.Path | str | None = None, api_key: str | None = None) -> None:
        super().__init__(cache_dir=cache_dir)
        if not isinstance(api_key, str):
            api_key = os.getenv("OPENAI_API_KEY")
            if not isinstance(api_key, str):
                raise ValueError("OpenAI API key is not provided")
        self.api_key = api_key

        self.client = openai.OpenAI(api_key=self.api_key)
    
    @property
    def service_name(self) -> str:
        return "OpenAI"


class OpenAITTSService(base.TTSService, OpenAIService):
    def __init__(
            self,
            voice: str = "alloy",
            model: str = "tts-1-hd",
            speed: float = 1.0,
            *,
            cache_dir: pathlib.Path | str | None = None,
            api_key: str | None = None
    ) -> None:
        super().__init__(cache_dir=cache_dir, api_key=api_key)
        self.voice = voice
        self.model = model
        self.speed = speed
    
    def tts(self, input: base.TTSInput) -> base.TTSData:
        file_name = self.get_file_name(input)

        data = base.TTSData(
            info=base.ServiceInfo(
                service_name=self.service_name,
                service_type=self.service_type,
                config={
                    "voice": self.voice,
                    "model": self.model,
                    "speed": self.speed
                }
            ),
            input=input,
            output=base.TTSOutput(audio_path=file_name)
        )

        if not (self.cache_dir / file_name).exists():
            response = self.client.audio.speech.create(
                input=input.text,
                model=self.model,
                voice=self.voice,
                speed=self.speed
            )
            response.stream_to_file(self.cache_dir / file_name)
        
        return data


class OpenAISTTService(base.STTService, OpenAIService):
    def __init__(
            self,
            model: str = "whisper-1",
            language: str | None = None,
            *,
            cache_dir: pathlib.Path | str | None = None,
            api_key: str | None = None
    ) -> None:
        super().__init__(cache_dir=cache_dir, api_key=api_key)
        self.model = model
        self.language = language
    
    def stt(self, input: base.STTInput) -> base.STTData:
        info = base.ServiceInfo(
            service_name=self.service_name,
            service_type=self.service_type,
            config={"model": self.model}
        )

        if isinstance(self.language, str):
            response = self.client.audio.transcriptions.create(
                file=(self.cache_dir / input.audio_path).open("rb"),
                model=self.model,
                language=self.language,
                response_format="verbose_json",
                timestamp_granularities=["word"]
            )
        else:
            response = self.client.audio.transcriptions.create(
                file=(self.cache_dir / input.audio_path).open("rb"),
                model=self.model,
                response_format="verbose_json",
                timestamp_granularities=["word"]
            )

        word_boundaries: list[base.Boundary] = []
        text_offset = 0
        for word in response.words:
            word_boundaries.append(
                base.Boundary(
                    text=word["word"],
                    start=word["start"],
                    end=word["end"],
                    text_offset=text_offset
                )
            )
            if text_offset != 0 and response.text[text_offset] == " ":
                text_offset += 1
            text_offset += len(word["word"])
        
        return base.STTData(info=info, input=input, output=base.STTOutput(text=response.text, boundaries=word_boundaries))


class OpenAITranslationService(base.TranslationService, OpenAIService):
    def __init__(self, model: str = "gpt-4o", *, cache_dir: pathlib.Path | str | None = None, api_key: str | None = None) -> None:
        super().__init__(cache_dir=cache_dir, api_key=api_key)
        self.model = model
        self.system_message = """Translate the given text from {source_language} to {target_language}. Do not output anything other than the translated text.
        If you encounter XML tags, do not translate their contents and insert them appropriately in the translated text. Do NOT skip any XML tags."""
    
    def translate(self, input: base.TranslationInput) -> base.TranslationData:
        info = base.ServiceInfo(
            service_name=self.service_name,
            service_type=self.service_type,
            config={
                "model": self.model
            }
        )
        result = self.client.chat.completions.create(
            model=self.model,
            messages=[
                {"role": "system", "content": self.system_message.format(source_language=input.source_language, target_language=input.target_language)},
                {"role": "user", "content": input.text}
            ],
            max_tokens=4095
        ).choices[0].message.content
        if not isinstance(result, str):
            raise ValueError(f"Unexpected response: {result}")
        return base.TranslationData(info=info, input=input, output=base.TranslationOutput(translated_text=result))



"""DeepL services."""

from . import base

import os
import pathlib
from abc import ABC, abstractmethod

try:
    import deepl
except ImportError:
    raise ImportError("Please install deepl with `pip install deepl`")


class DeepLService(base.Service, ABC):
    def __init__(self, *, cache_dir: pathlib.Path | str | None = None, api_key: str | None = None) -> None:
        super().__init__(cache_dir=cache_dir)
        if not isinstance(api_key, str):
            api_key = os.getenv("DEEPL_API_KEY")
            if not isinstance(api_key, str):
                raise ValueError("DeepL API key is not provided")
        self.api_key = api_key
    
    @property
    def service_name(self) -> str:
        return "DeepL"


class DeepLTranslationService(base.TranslationService, DeepLService):
    def __init__(self, *, cache_dir: pathlib.Path | str | None = None, api_key: str | None = None) -> None:
        super().__init__(cache_dir=cache_dir, api_key=api_key)
        self.client = deepl.Translator(auth_key=self.api_key)

    def translate(self, input: base.TranslationInput) -> base.TranslationData:
        info = base.ServiceInfo(
            service_name=self.service_name,
            service_type=self.service_type,
            config={}
        )
        if input.target_language == "zh_tw":
            try:
                import opencc
            except ImportError:
                raise ImportError("Please install opencc with `pip install opencc`")
            result = self.client.translate_text(input.text, source_lang=input.source_language, target_lang="zh", tag_handling="xml")
            return base.TranslationData(info=info, input=input, output=base.TranslationOutput(translated_text=opencc.OpenCC("s2t.json").convert(result.text)))
        target_language = input.target_language
        if target_language == "zh_cn":
            target_language = "zh"
        result = self.client.translate_text(input.text, source_lang=input.source_language, target_lang=target_language, tag_handling="xml")
        return base.TranslationData(info=info, input=input, output=base.TranslationOutput(translated_text=result.text))



"""ElevenLabs services."""

from . import base

import os
import pathlib
from abc import ABC, abstractmethod

try:
    import elevenlabs.client
except ImportError:
    raise ImportError("Please install elevenlabs with `pip install elevenlabs`")


class ElevenLabsService(base.Service, ABC):
    def __init__(self, *, cache_dir: pathlib.Path | str | None = None, api_key: str | None = None) -> None:
        super().__init__(cache_dir=cache_dir)
        if not isinstance(api_key, str):
            api_key = os.getenv("ELEVEN_API_KEY")
            if not isinstance(api_key, str):
                raise ValueError("ElevenLabs API key is not provided")
        self.api_key = api_key

        self.client = elevenlabs.client.ElevenLabs(api_key=self.api_key)
    
    @property
    def service_name(self) -> str:
        return "ElevenLabs"


class ElevenLabsTTSService(base.TTSService, ElevenLabsService):
    def __init__(
            self,
            voice: str,
            model: str = "eleven_multilingual_v2",
            *,
            cache_dir: pathlib.Path | str | None = None,
            api_key: str | None = None
    ) -> None:
        super().__init__(cache_dir=cache_dir, api_key=api_key)
        self.voice = voice
        self.model = model

    def tts(self, input: base.TTSInput) -> base.TTSData:
        file_name = self.get_file_name(input)

        data = base.TTSData(
            info=base.ServiceInfo(
                service_name=self.service_name,
                service_type=self.service_type,
                config={
                    "voice": self.voice,
                    "model": self.model
                }
            ),
            input=input,
            output=base.TTSOutput(audio_path=file_name)
        )

        if not (self.cache_dir / file_name).exists():
            response = self.client.generate(
                text=input.text,
                voice=self.voice,
                model=self.model
            )
            elevenlabs.save(response, str(self.cache_dir / file_name))
        
        return data



"""Text translation functions for Manim Speech."""

from . import services

import subprocess
import pathlib

import polib
import manim


def init_translation_env(file: pathlib.Path | str, domain: str) -> None:
    if not pathlib.Path("locales").exists():
        pathlib.Path("locales").mkdir()
    result = subprocess.run(["xgettext", "-d", domain, "-o", str(pathlib.Path("locales") / f"{domain}.pot"), str(file)])
    if result.returncode != 0:
        raise RuntimeError(f"xgettext failed with return code {result.returncode}")


def translate_po_file(domain: str, src_lang: str, target_lang: str, *, service: services.base.TranslationService | None = None) -> None:
    src_path = pathlib.Path("locales") / f"{domain}.pot"
    target_path = pathlib.Path("locales") / target_lang / "LC_MESSAGES" / f"{domain}"
    if not target_path.parent.exists():
        target_path.parent.mkdir(parents=True)
    manim.logger.info(f"Translating to {target_lang}...")
    if not target_path.with_suffix(".po").exists():
        pofile = polib.pofile(str(src_path))
        pofile.metadata["Content-Type"] = "text/plain; charset=UTF-8"
        if isinstance(service, services.base.TranslationService):
            manim.logger.info(f"Using {service.service_name} translation service.")
            for entry in pofile.untranslated_entries():
                translation = service.translate(services.base.TranslationInput(text=entry.msgid, source_language=src_lang, target_language=target_lang)).output.translated_text
                entry.msgstr = translation
        else:
            manim.logger.info("No translation service specified.")
        pofile.save(str(target_path.with_suffix(".po")))
        if not isinstance(service, services.base.TranslationService):
            manim.console.print(f"An empty translation file has been created at {target_path.with_suffix(".po")}. Please fill it in and then rerun `manim`.")
            exit(1)
    else:
        manim.logger.info(f"Translation file for {target_lang} found.")
        pofile = polib.pofile(str(target_path.with_suffix(".po")))
    pofile.save_as_mofile(str(target_path.with_suffix(".mo")))



"""AssemblyAI services."""

from . import base

import os
import typing
import pathlib
from abc import ABC, abstractmethod

try:
    import assemblyai as aai
except ImportError:
    raise ImportError("Please install assemblyai with `pip install assemblyai`")


class AssemblyAIService(base.Service, ABC):
    def __init__(self, *, cache_dir: pathlib.Path | str | None = None, api_key: str | None = None) -> None:
        super().__init__(cache_dir=cache_dir)
        if not isinstance(api_key, str):
            api_key = os.getenv("ASSEMBLYAI_API_KEY")
            if not isinstance(api_key, str):
                raise ValueError("AssemblyAI API key is not provided")
        self.api_key = api_key
    
    @property
    def service_name(self) -> str:
        return "AssemblyAI"


class AssemblyAISTTService(base.STTService, AssemblyAIService):
    def __init__(
            self,
            model: typing.Literal["best", "nano"] = "best",
            language: str | None = None,
            word_boost: list[str] | None = None,
            custom_spelling: dict[str, str | list[str]] | None = None,
            *,
            cache_dir: pathlib.Path | str | None = None,
            api_key: str | None = None
    ) -> None:
        super().__init__(cache_dir=cache_dir, api_key=api_key)
        self.model = model
        self.language = language
        self.word_boost = word_boost if isinstance(word_boost, list) else []
        self.custom_spelling = custom_spelling if isinstance(custom_spelling, dict) else {}

        self.config = aai.TranscriptionConfig(
            speech_model=aai.SpeechModel.best if model == "best" else aai.SpeechModel.nano,
            language_code=language if isinstance(language, str) else None,
            language_detection=isinstance(language, type(None)),
            word_boost=self.word_boost,
            custom_spelling=self.custom_spelling,
            punctuate=False
        )

    def stt(self, input: base.STTInput) -> base.STTData:
        info = base.ServiceInfo(
            service_name=self.service_name,
            service_type=self.service_type,
            config={
                "model": self.model,
                "language": self.language,
                "word_boost": self.word_boost,
                "custom_spelling": self.custom_spelling
            }
        )

        aai.settings.api_key = self.api_key
        transcriber = aai.Transcriber(config=self.config)
        response = transcriber.transcribe(str(self.cache_dir / input.audio_path))
        if response.error:
            raise ValueError(response.error)

        word_boundaries: list[base.Boundary] = []
        text_offset = 0
        for word in response.words:
            word_boundaries.append(
                base.Boundary(
                    text=word.text,
                    start=word.start / 1000,
                    end=word.end / 1000,
                    text_offset=text_offset
                )
            )
            if text_offset != 0 and response.text[text_offset] == " ":
                text_offset += 1
            text_offset += len(word.text)
    
        return base.STTData(
            info=info,
            input=input,
            output=base.STTOutput(text=response.text, boundaries=word_boundaries)
        )



"""Speech utils for Manim Speech."""

from . import services

import pathlib
import re

import manim
from scipy import interpolate
import pydantic
from mutagen.mp3 import MP3


class SpeechData(pydantic.BaseModel):
    text: str
    tts_data: services.base.TTSData
    stt_data: services.base.STTData
    duration: float
    bookmark_times: dict[str, float]


def remove_bookmarks(s: str) -> str:
    return re.sub(r"<bookmark\s*mark\s*=['\"]\w*[\"']\s*/>", "", s)


def get_bookmark_times(text: str, tts_data: services.base.TTSData, stt_data: services.base.STTData) -> dict[str, float]:
    interpolator = interpolate.interp1d(
        [wb.text_offset for wb in stt_data.output.boundaries] + [len(stt_data.output.text)],
        [wb.start for wb in stt_data.output.boundaries] + [stt_data.output.boundaries[-1].end],
        fill_value="extrapolate"
    )

    text_len = len(tts_data.input.text)
    transcribed_text_len = len(stt_data.output.text.strip())

    bookmark_dist: dict[str, int] = {}
    content: str = ""
    for part in re.split(r"(<bookmark\s*mark\s*=[\'\"]\w*[\"\']\s*/>)", text):
        match = re.match(r"<bookmark\s*mark\s*=[\'\"](.*)[\"\']\s*/>", part)
        if not isinstance(match, type(None)):
            bookmark_dist[match.group(1)] = len(content)
        else:
            content += part
        
    bookmark_times: dict[str, float] = {}
    for mark, dist in bookmark_dist.items():
        bookmark_times[mark] = interpolator(dist * transcribed_text_len / text_len)

    return bookmark_times


def load_cache(cache_dir: pathlib.Path) -> list[SpeechData]:
    with (cache_dir / "cache.json").open("r") as f:
        return pydantic.TypeAdapter(list[SpeechData]).validate_json(f.read())


def save_cache(cache_dir: pathlib.Path, data: list[SpeechData]) -> None:
    if not cache_dir.exists():
        cache_dir.mkdir(parents=True)
    with (cache_dir / "cache.json").open("w") as f:
        f.write(pydantic.TypeAdapter(list[SpeechData]).dump_json(data, indent=4).decode("utf-8"))


def create(
        text: str,
        tts_service: services.base.TTSService | None = None,
        stt_service: services.base.STTService | None = None,
        *,
        cache_dir: pathlib.Path | str | None = None
) -> SpeechData:
    if isinstance(cache_dir, type(None)):
        cache_dir = pathlib.Path(manim.config.media_dir) / "manim_speech"
    elif isinstance(cache_dir, str):
        cache_dir = pathlib.Path(cache_dir)

    if not cache_dir.exists():
        cache_dir.mkdir(parents=True)
    
    if (cache_dir / "cache.json").exists():
        cache = load_cache(cache_dir)
        for data in cache:
            if data.text == text:
                return data
    else:
        cache = []

    input_text = remove_bookmarks(text)
    if isinstance(tts_service, services.base.TTSService):
        tts_data = tts_service.tts(services.base.TTSInput(text=input_text))
    else:
        input_data = services.base.TTSInput(text=input_text)
        tts_data = services.base.TTSData(
            info=None,
            input=input_data,
            output=services.base.TTSOutput(audio_path=services.base.TTSService.get_file_name(input_data))
        )
        if not (cache_dir / tts_data.output.audio_path).exists():
            manim.console.print(f"Please record the following text manually and save it to {cache_dir / tts_data.output.audio_path}.")
            manim.console.print(f"[green]{input_text}[/green]")
            manim.console.print("Rerun `manim` after you're done recording.")
            exit(1)
    if isinstance(stt_service, services.base.STTService):
        stt_data = stt_service.stt(services.base.STTInput(audio_path=tts_data.output.audio_path))
    else:
        stt_data = services.base.STTData(
            info=None,
            input=services.base.STTInput(audio_path=tts_data.output.audio_path),
            output=services.base.STTOutput(
                text=input_text,
                boundaries=[
                    services.base.Boundary(
                        text=input_text,
                        start=0.0,
                        end=MP3(cache_dir / tts_data.output.audio_path).info.length,
                        text_offset=0
                    )
                ]
            )
        )
    bookmark_times = get_bookmark_times(text, tts_data, stt_data)

    data = SpeechData(
        text=text,
        tts_data=tts_data,
        stt_data=stt_data,
        duration=MP3(cache_dir / tts_data.output.audio_path).info.length,
        bookmark_times=bookmark_times,
    )
    cache.append(data)
    save_cache(cache_dir, cache)

    return data



"""Base classes for services."""

import pathlib
import hashlib
from abc import ABC, abstractmethod

import manim
import pydantic


class ServiceInfo(pydantic.BaseModel):
    service_name: str
    service_type: str
    config: pydantic.JsonValue


class Service(ABC):
    def __init__(self, *, cache_dir: pathlib.Path | str | None = None) -> None:
        if isinstance(cache_dir, type(None)):
            cache_dir = pathlib.Path(manim.config.media_dir) / "manim_speech"
        elif isinstance(cache_dir, str):
            cache_dir = pathlib.Path(cache_dir)
        self.cache_dir = cache_dir

        if not self.cache_dir.exists():
            self.cache_dir.mkdir(parents=True)
    
    @property
    @abstractmethod
    def service_name(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def service_type(self) -> str:
        raise NotImplementedError


class TTSInput(pydantic.BaseModel):
    text: str


class TTSOutput(pydantic.BaseModel):
    audio_path: pathlib.Path


class TTSData(pydantic.BaseModel):
    info: ServiceInfo | None
    input: TTSInput
    output: TTSOutput


class TTSService(Service, ABC):
    @property
    def service_type(self) -> str:
        return "TTS"
    
    @staticmethod
    def get_file_name(input: TTSInput) -> pathlib.Path:
        return pathlib.Path(f"{hashlib.sha256(input.text.encode()).hexdigest()}.mp3")
    
    @abstractmethod
    def tts(self, input: TTSInput) -> TTSData:
        raise NotImplementedError


class Boundary(pydantic.BaseModel):
    text: str
    start: float
    end: float
    text_offset: int

    @property
    def length(self) -> int:
        return len(self.text)


class STTInput(pydantic.BaseModel):
    audio_path: pathlib.Path


class STTOutput(pydantic.BaseModel):
    text: str
    boundaries: list[Boundary]


class STTData(pydantic.BaseModel):
    info: ServiceInfo | None
    input: STTInput
    output: STTOutput


class STTService(Service, ABC):
    @property
    def service_type(self) -> str:
        return "STT"
    
    @abstractmethod
    def stt(self, input: STTInput) -> STTData:
        raise NotImplementedError


class TranslationInput(pydantic.BaseModel):
    text: str
    source_language: str
    target_language: str


class TranslationOutput(pydantic.BaseModel):
    translated_text: str


class TranslationData(pydantic.BaseModel):
    info: ServiceInfo | None
    input: TranslationInput
    output: TranslationOutput


class TranslationService(Service, ABC):
    @property
    def service_type(self) -> str:
        return "Translation"
    
    @abstractmethod
    def translate(self, input: TranslationInput) -> TranslationData:
        raise NotImplementedError
